#acquirefile.py: reads *.acquire data files.  By Chris Nicolai.  Thanks to Bruxton for sharing AcquireFormat.h.

#// AcquireFormat.h
#//
#// Copyright (c) 1999 Bruxton Corporation
#//
#// Author: Joseph Grote
#
#// Bruxton Corporation Acquire data format.
#
#// The file is written with Intel byte ordering.
#
#// This file format is designed so that all values are 8 byte aligned.
#
#// Data is always written to the file in multiples of ACQUIRE_BLOCK_SIZE.
#
#// The data file has a hierarchical structure, organized as series, sweeps
#// and segments. Within segments there are channels and data blocks.
#
#// The file format is as follows:
#
#//*****************************************************************
#// header
#// file prefix block
#// sweep prefix block
#// data block
#// data block
#//		.
#//		.
#// sweep block
#// sweep prefix block
#// data block
#//		.
#//		.
#// series block
#// sweep block
#//		.
#//		.
#//		.
#// series block
#// control block
#// header
#// AcquirePrefix
#//*****************************************************************
#
#// the header contains 
#// AcquirePrefix
#// AcquireHeader
#
#// AcquirePrefix is duplicated at the end of the file in an ACQUIRE_BLOCK_SIZE size block
#// in order to be consistent with previous versions of the file format.
#// DataAccess first checks the the last block to determine the file type and version.
#
#// The file prefix block contains:
#
#// array of channel records, grouped in the following order:
#
#// AcquireAnalogInChannelRecords
#// AcquireAperiodicChannelRecords
#
#// A series block contains:
#
#// AcquireSeriesDescriptor
#// directory of sweep prefix block offsets within the control block.
#
#// a sweep prefix block contains:
#
#// array of analog_in_channel_count AcquireSweepPrefixAnalogChannel
#// array of aperiodic_channel_count AcquireSweepPrefixAperiodicChannel
#
#// a data block contains a block of data.
#
#// a sweep block contains:
#
#// AcquireSweepDescriptor
#// AcquireChannelDescriptor	for the control channels
#// AcquireBlockDescriptors for the control channels
#// AcquireChannelDescriptor	for the control channel extended data
#// AcquireBlockDescriptors for the control channel extended data
#// Segment directory listing AcquireSegmentDescriptor offsets within the control block
#// AcquireSegmentDescriptor
#// Channel directory listing AcquireChannelDescriptor offsets within the control block
#//    for the analog input channels
#// AcquireChannelDescriptor
#// AcquireBlockDescriptors
#//		.
#//		.
#//		.
#// AcquireSegmentDescriptor
#// Channel directory
#// AcquireChannelDescriptor
#// AcquireBlockDescriptors
#//		.
#//		.
#//		.
#
#// the control block contains:
#
#// array of channel records, grouped in the following order:
#
#// AcquireAnalogInChannelRecords
#// AcquireAperiodicChannelRecords
#
#// repeat of series, sweep prefix and sweep blocks (without individual block tags
#// and block boundary padding).
#// a directory of series block offsets within the control block
#
#// V6.1 Set data_size in tags for all block types except the header blocks.
#

import mmap
import numpy
import os
import struct
import traceback
from math import *

ACQUIRE_FORMAT = 1
ACQUIRE_BLOCK_SIZE = 1024
ACQUIRE_FILE_COMMENT_SIZE = 256
ACQUIRE_COMMENT_SIZE = 80
ACQUIRE_FILE_TECHNIQUE_SIZE = 256
ACQUIRE_TECHNIQUE_SIZE = 80
ACQUIRE_SCREEN_SIZE = 80
ACQUIRE_LABEL_SIZE = 16
ACQUIRE_UNITS_SIZE = 16
ACQUIRE_SIGNAL_SIZE = 32
ACQUIRE_CONTROL_NAME_SIZE = 80
ACQUIRE_USER_NAME_SIZE = 64
ACQUIRE_COMPUTER_NAME_SIZE = 64

try:
    from qubx.fast.data import decode_acquirefile_samples
except:
    def decode_acquirefile_samples(src_8, src_16, src_16s, dest, size, count):
        i_dest = 0
        i_src = 0
        val = 0
        while i_src < size:
            c = src[i_src]
            if c == -128:
                i_src += 1
                if i_src % 2:
                    val = src_16s[i_src/2]
                else:
                    val = src_16[i_src/2]
                i_src += 2
            else:
                val += c
                i_src += 1
            dest[i_dest] = val
            i_dest += 1



def read_string8(buf, loc_bytes, count):
    next = loc_bytes + count
    s = str(buf[loc_bytes:next])
    terminator = s.find(chr(0))
    if terminator >= 0:
        s = s[:terminator]
    return s, next

def read_string16(buf, loc_bytes, count):
    next = loc_bytes+2*count
    s = unicode(buf[loc_bytes:next], encoding='UTF-16')
    terminator = s.find(chr(0))
    if terminator >= 0:
        s = s[:terminator]
    return s, next

def read_struct(buf, loc_bytes, format):
    bytes = struct.calcsize(format)
    next = loc_bytes+bytes
    return list(struct.unpack(format, buf[loc_bytes:next])) + [next]

#
#// AcquirePrefix contains information to go at the head of the file.
#// It contains the same fields for all file format versions.
#
#// Note that Acquire file format version 1 uses the reserved space
#// and is in Macintosh byte ordering.
#// See SKALARFormatV1.h.
#
#struct AcquirePrefix
#{
#	int8_t vendor[8];		// company name
#	int32_t format;			// application
#	int32_t version;		// version number
#	int64_t minor_version;
#	int64_t reserved[3];
#};
class AcquirePrefix(object):
    def __init__(self, buf, loc_bytes):
        self.loc_bytes = loc_bytes
        self.vendor, loc_bytes = read_string8(buf, loc_bytes, 8)
        self.format, self.version, self.minor_version, r1, r2, r3, self.next_loc_bytes = read_struct(buf, loc_bytes, 'iiqqqq')
        self.reserved = (r1, r2, r3)
#
##ifndef __cplusplus
#typedef struct AcquirePrefix AcquirePrefix;
##endif
#
#// AcquireHeader
#
#struct AcquireHeader
#{
#
#	int64_t date_time;			// date/time created in 100 ns units from Jan 1 1601
#	int64_t file_tag;			// file tag used for reconstruction of damaged data
#									// use the same value in mBlockTag				
#	int16_t comment[ACQUIRE_FILE_COMMENT_SIZE];
#	int16_t technique[ACQUIRE_FILE_TECHNIQUE_SIZE];
#	double clock_rate;				// base clock rate in Hz
#	int64_t frame_interval;		// frame interval in clock ticks
#	int64_t analog_in_channel_count;
#	int64_t analog_in_maximum_block_size;	// maximum possible size of a data block in bytes
#	int64_t analog_in_maximum_data_count; 	// maximum data values per data block
#	int64_t aperiodic_channel_count;
#	int64_t aperiodic_maximum_block_size;	// maximum possible size of a data block in bytes
#	int64_t series_count;
#	int64_t file_size;			// file size in bytes
#	int64_t control_size;		// size of the control block in bytes
#	int64_t control_position;		// position of control block in the file
#	int64_t directory_position;		// series directory offset in control block
#	int16_t user_name[ACQUIRE_USER_NAME_SIZE];	// Current logged in user.
#	int16_t computer_name[ACQUIRE_USER_NAME_SIZE];	// Current logged in user.
#};
class AcquireHeader(object):
    def __init__(self, buf, loc_bytes):
        self.loc_bytes = loc_bytes
        self.date_time, self.file_tag, loc_bytes = read_struct(buf, loc_bytes, 'qq')
        self.comment, loc_bytes = read_string16(buf, loc_bytes, ACQUIRE_FILE_COMMENT_SIZE)
        self.technique, loc_bytes = read_string16(buf, loc_bytes, ACQUIRE_FILE_TECHNIQUE_SIZE)
        (self.clock_rate,
         self.frame_interval,
         self.analog_in_channel_count,
         self.analog_in_maximum_block_size,
         self.analog_in_maximum_data_count,
         self.aperiodic_channel_count,
         self.aperiodic_maximum_block_size,
         self.series_count,
         self.file_size,
         self.control_size,
         self.control_position,
         self.directory_position,
         loc_bytes) = read_struct(buf, loc_bytes, 'dqqqqqqqqqqq')
        self.user_name, loc_bytes = read_string16(buf, loc_bytes, ACQUIRE_USER_NAME_SIZE)
        self.computer_name, self.next_loc_bytes = read_string16(buf, loc_bytes, ACQUIRE_USER_NAME_SIZE)
#
##ifndef __cplusplus
#typedef struct AcquireHeader AcquireHeader;
##endif
#
#// AcquireSeriesDescriptor
#
#struct AcquireSeriesDescriptor
#{
#	int16_t comment[ACQUIRE_COMMENT_SIZE];
#	int16_t technique[ACQUIRE_TECHNIQUE_SIZE];
#	int64_t start_time;		// time of start of series 100 ns units from Jan 1 1601 
#	int64_t sweep_count;
#};
class AcquireSeriesDescriptor(object):
    def __init__(self, buf, loc_bytes):
        self.loc_bytes = loc_bytes
        self.comment, loc_bytes = read_string16(buf, loc_bytes, ACQUIRE_COMMENT_SIZE)
        self.technique, loc_bytes = read_string16(buf, loc_bytes, ACQUIRE_TECHNIQUE_SIZE)
        self.start_time, self.sweep_count, self.next_loc_bytes = read_struct(buf, loc_bytes, 'qq')
#
##ifndef __cplusplus
#typedef struct AcquireSeriesDescriptor AcquireSeriesDescriptor;
##endif
#
#// AcquireSweepPrefixAnalogChannel
#
#// Channel sweep prefix for analog channels.
#// Gain is the conversion from V to the SI units representing an external amplifier.
#// In version 6.0 offset is a place holder for potential amplifiers that supply an offset.
#// It should always be 0.
#
#struct AcquireSweepPrefixAnalogChannel
#{	
#	double maximum_voltage;
#	double minimum_voltage;
#	int64_t maximum_value;
#	int64_t minimum_value;
#	double gain;
#	double offset;
#};
class AcquireSweepPrefixAnalogChannel(object):
    def __init__(self, buf, loc_bytes):
        self.loc_bytes = loc_bytes
        (self.maximum_voltage, self.minimum_voltage,
         self.maximum_value, self.minimum_value,
         self.gain, self.offset,
         self.next_loc_bytes) = read_struct(buf, loc_bytes, 'ddqqdd')
#
##ifndef __cplusplus
#typedef struct AcquireSweepPrefixAnalogChannel AcquireSweepPrefixAnalogChannel;
##endif
#
#// AcquireSweepPrefixAperiodicChannel
#
#// Channel sweep prefix for control channels.
#// value is the control holding value.
#
#struct AcquireSweepPrefixAperiodicChannel
#{	
#	double value;
#	int64_t reserved[4];
#};
class AcquireSweepPrefixAperiodicChannel(object):
    def __init__(self, buf, loc_bytes):
        self.loc_bytes = loc_bytes
        self.value, r1, r2, r3, r4, self.next_loc_bytes = read_struct(buf, loc_bytes, 'dqqqq')
#
##ifndef __cplusplus
#typedef struct AcquireSweepPrefixAperiodicChannel AcquireSweepPrefixAperiodicChannel;
##endif
#
#// AcquireSweepDescriptor 
#
#struct AcquireSweepDescriptor
#{
#	int16_t comment[ACQUIRE_COMMENT_SIZE];
#	int16_t technique[ACQUIRE_TECHNIQUE_SIZE];
#	int16_t screen[ACQUIRE_SCREEN_SIZE];
#	int64_t start_time;		// time of start of sweep 100 ns units from Jan 1 1601 
#	int64_t segment_count;
#	int64_t duration;		// sweep duration in frames
#	int64_t control_data_location;	// Location of the control data channel descriptor in the control block.
#	int64_t segment_location;	// Location of the segment directory in the control block.
#};
class AcquireSweepDescriptor(object):
    def __init__(self, buf, loc_bytes):
        self.loc_bytes = loc_bytes
        self.comment, loc_bytes = read_string16(buf, loc_bytes, ACQUIRE_COMMENT_SIZE)
        self.technique, loc_bytes = read_string16(buf, loc_bytes, ACQUIRE_TECHNIQUE_SIZE)
        self.screen, loc_bytes = read_string16(buf, loc_bytes, ACQUIRE_SCREEN_SIZE)
        (self.start_time, self.segment_count,
         self.duration, self.control_data_location,
         self.segment_location, self.next_loc_bytes) = read_struct(buf, loc_bytes, 'qqqqq')
#
##ifndef __cplusplus
#typedef struct AcquireSweepDescriptor AcquireSweepDescriptor;
##endif
#
#// AcquireSegmentDescriptor
#
#struct AcquireSegmentDescriptor
#{
#	int64_t start;		// start time in frames from the start of the sweep
#	int64_t duration;	// segment duration in frames
#};
class AcquireSegmentDescriptor(object):
    def __init__(self, buf, loc_bytes):
        self.loc_bytes = loc_bytes
        self.start, self.duration, self.next_loc_bytes = read_struct(buf, loc_bytes, 'qq')
#
##ifndef __cplusplus
#typedef struct AcquireSegmentDescriptor AcquireSegmentDescriptor;
##endif
#
#// AcquireChannelDescriptor
#
#struct AcquireChannelDescriptor
#{
#	int64_t block_count;
#};
class AcquireChannelDescriptor(object):
    def __init__(self, buf, loc_bytes):
        self.loc_bytes = loc_bytes
        self.block_count, self.next_loc_bytes = read_struct(buf, loc_bytes, 'q')
#
##ifndef __cplusplus
#typedef struct AcquireChannelDescriptor AcquireChannelDescriptor;
##endif
#
#// AcquireBlockDescriptor
#
#struct AcquireBlockDescriptor
#{
#	int64_t size;		// block size in bytes
#	int64_t data_count;
#	int64_t data_size;  // data size in bytes
#	int64_t location;	// data block location within the file
#};
class AcquireBlockDescriptor(object):
    def __init__(self, buf, loc_bytes):
        self.loc_bytes = loc_bytes
        self.size, self.data_count, self.data_size, self.location, self.next_loc_bytes = read_struct(buf, loc_bytes, 'qqqq')
#
##ifndef __cplusplus
#typedef struct AcquireBlockDescriptor AcquireBlockDescriptor;
##endif
#
#// AcquireAnalogInChannelRecord
#
#// Channel record for analog input channels.
#
#// natural_scale is the power of 1000 to multiply the units to get the natural units.
#// E.g., if units are A and the natural units are pA, natural_scale is -4.
#
#struct AcquireAnalogInChannelRecord
#{
#	int16_t label[ACQUIRE_LABEL_SIZE];	// physical channel identification in wide chars
#	int16_t signal[ACQUIRE_SIGNAL_SIZE];	// channel description in wide chars
#	int16_t units[ACQUIRE_UNITS_SIZE];	// SI units of the data.
#	int64_t natural_scale;	// natural_scale scale conversion.
#};
class AcquireAnalogInChannelRecord(object):
    def __init__(self, buf, loc_bytes):
        self.loc_bytes = loc_bytes
        self.label, loc_bytes = read_string16(buf, loc_bytes, ACQUIRE_LABEL_SIZE)
        self.signal, loc_bytes = read_string16(buf, loc_bytes, ACQUIRE_SIGNAL_SIZE)
        self.units, loc_bytes = read_string16(buf, loc_bytes, ACQUIRE_UNITS_SIZE)
        self.natural_scale, self.next_loc_bytes = read_struct(buf, loc_bytes, 'q')
#
##ifndef __cplusplus
#typedef struct AcquireAnalogInChannelRecord AcquireAnalogInChannelRecord;
##endif
#
#// AcquireAperiodicChannelRecord
#
#struct AcquireAperiodicChannelRecord
#{
#	int64_t type;
#	int16_t name[ACQUIRE_CONTROL_NAME_SIZE];
#	int16_t description[ACQUIRE_COMMENT_SIZE];
#	int16_t signal[ACQUIRE_SIGNAL_SIZE];
#	int16_t units[ACQUIRE_UNITS_SIZE];
#	int64_t natural_scale;	// natural_scale scale conversion.
#	int64_t reserved[8];
#};
class AcquireAperiodicChannelRecord(object):
    def __init__(self, buf, loc_bytes):
        self.loc_bytes = loc_bytes
        self.type, loc_bytes = read_struct(buf, loc_bytes, 'q')
        self.name, loc_bytes = read_string16(buf, loc_bytes, ACQUIRE_CONTROL_NAME_SIZE)
        self.description, loc_bytes = read_string16(buf, loc_bytes, ACQUIRE_COMMENT_SIZE)
        self.signal, loc_bytes = read_string16(buf, loc_bytes, ACQUIRE_SIGNAL_SIZE)
        self.units, loc_bytes = read_string16(buf, loc_bytes, ACQUIRE_UNITS_SIZE)
        ints = read_struct(buf, loc_bytes, '9q')
        self.natural_scale = ints[0]
        self.reserved = ints[1:-1]
        self.next_loc_bytes = ints[-1]
#
##ifndef __cplusplus
#typedef struct AcquireAperiodicChannelRecord AcquireAperiodicChannelRecord;
##endif
#
#// Aperiodic channel types.
#
#enum
#{
ACQUIRE_APERIODIC_CHANNEL_CONTROL = 0
ACQUIRE_APERIODIC_CHANNEL_IMAGE = 1
#};
#
#// AcquireAperiodicEntry
#
#// Entry for an aperidic channel.
#// The entry may contain an attached text string, data and extended data.
#// The int16_t text string immediately follows the marker in memory.
#// The text string is zero terminated.
#// The data immediately follows the text in memory.
#// The extended data is stored as a block in the extended data channel.
#
#struct AcquireAperiodicEntry
#{
#	int64_t entry_size;		// size in bytes including the text string and data.
#	int64_t control_channel;
#	int64_t frame;
#	double value;
#	int64_t text_size;	// size of the text string in characters including the terminating zero.
#	int64_t data_size;	// size of the data in bytes.
#	int64_t extended_data_size;	// size of the extended data in bytes.
#	int64_t extended_block_index; // block index of the extended data.
#};
class AcquireAperiodicEntry(object):
    def __init__(self, buf, loc_bytes):
        self.loc_bytes = loc_bytes
        (self.entry_size, self.control_channel,
         self.frame, self.value,
         self.text_size, self.data_size,
         self.extended_data_size, self.extended_block_index,
         self.next_loc_bytes) = read_struct(buf, loc_bytes, 'qqqdqqqq')
#
##ifndef __cplusplus
#typedef struct AcquireAperiodicEntry AcquireAperiodicEntry;
##endif
#
#// mBlockTag
#
#// A tag to prepend all blocks to be used in reconstruction
#// of damaged data files.
#
#struct AcquireBlockTag
#{
#	int64_t file_tag;		// same as AcquireHeader.file_tag
#	int64_t block_index;
#	int64_t block_type;		// block type, enumerated below
#	int64_t block_size;		// block size in bytes
#	int64_t data_count;
#	int64_t data_size;		// data size in bytes
#	int64_t	channel;
#	int64_t frame;
#};
class AcquireBlockTag(object):
    def __init__(self, buf, loc_bytes):
        self.loc_bytes = loc_bytes
        (self.file_tag, self.block_index,
         self.block_type, self.block_size,
         self.data_count, self.data_size,
         self.channel, self.frame,
         self.next_loc_bytes) = read_struct(buf, loc_bytes, 'qqqqqqqq')
#
##ifndef __cplusplus
#typedef struct AcquireBlockTag AcquireBlockTag;
##endif
#
#// Block types.
#
#enum
#
ACQUIRE_DATA_BLOCK = 1
ACQUIRE_SWEEP_BLOCK = 2
ACQUIRE_CONTROL_BLOCK = 3
ACQUIRE_SERIES_BLOCK = 4
ACQUIRE_FILE_PREFIX_BLOCK = 5
ACQUIRE_SWEEP_PREFIX_BLOCK = 6
#;
#
##endif

def round_up(i, divisor):
    return divisor * int(ceil(i*1.0/divisor))

def ReadHeader(buf):
    prefix = AcquirePrefix(buf, 0)
    if prefix.format != ACQUIRE_FORMAT:
        raise IOError('Unknown format or invalid file: format %i' % prefix.format)
    if not (2 <= prefix.version <= 6):
        raise IOError('Unknown version or invalid file: version %i' % prefix.version)
    header = AcquireHeader(buf, prefix.next_loc_bytes)
    header.prefix = prefix
    return header, round_up(header.next_loc_bytes, ACQUIRE_BLOCK_SIZE)

class Block(object):
    def __init__(self, tag):
        self.tag = tag

def ReadFilePrefix(buf, loc_bytes, block, header):
    ai_ch, ap_ch = [], []
    for i in xrange(header.analog_in_channel_count):
        chan = AcquireAnalogInChannelRecord(buf, loc_bytes)
        ai_ch.append(chan)
        loc_bytes = chan.next_loc_bytes
    for i in xrange(header.aperiodic_channel_count):
        chan = AcquireAperiodicChannelRecord(buf, loc_bytes)
        ap_ch.append(chan)
        loc_bytes = chan.next_loc_bytes
    block.analog_in_channel_records = ai_ch
    block.aperiodic_channel_records = ap_ch

def ReadSweepPrefix(buf, loc_bytes, block, header):
    ai_ch_pre, ap_ch_pre = [], []
    for i in xrange(header.analog_in_channel_count):
        chan = AcquireSweepPrefixAnalogChannel(buf, loc_bytes)
        ai_ch_pre.append(chan)
        loc_bytes = chan.next_loc_bytes
    for i in xrange(header.aperiodic_channel_count):
        chan = AcquireSweepPrefixAperiodicChannel(buf, loc_bytes)
        ap_ch_pre.append(chan)
        loc_bytes = chan.next_loc_bytes
    block.analog_in_channel_prefixes = ai_ch_pre
    block.aperiodic_channel_prefixes = ap_ch_pre

def ReadSweep(buf, loc_bytes, block, header):
    sweep = AcquireSweepDescriptor(buf, loc_bytes)
    sweep.ctrl_channels = []
    ctrl_channels = AcquireChannelDescriptor(buf, sweep.next_loc_bytes)
    loc_bytes = ctrl_channels.next_loc_bytes
    for i in xrange(ctrl_channels.block_count):
        chan = AcquireBlockDescriptor(buf, loc_bytes)
        loc_bytes = chan.next_loc_bytes
        sweep.ctrl_channels.append(chan)
    sweep.ctrl_channels_ext = []
    ctrl_channels_ext = AcquireChannelDescriptor(buf, loc_bytes)
    loc_bytes = ctrl_channels_ext.next_loc_bytes
    for i in xrange(ctrl_channels_ext.block_count):
        chan = AcquireBlockDescriptor(buf, loc_bytes)
        loc_bytes = chan.next_loc_bytes
        sweep.ctrl_channels_ext.append(chan)
    ints = read_struct(buf, loc_bytes, '%iq'%sweep.segment_count)
    sweep.segment_dir, loc_bytes = ints[:-1], ints[-1]
    ints = read_struct(buf, loc_bytes, '%iq'%header.analog_in_channel_count)
    sweep.channel_dir, loc_bytes = ints[:-1], ints[-1]
    sweep.segments = []                
    for i in xrange(sweep.segment_count):
        seg = AcquireSegmentDescriptor(buf, loc_bytes)
        loc_bytes = seg.next_loc_bytes
        seg.channels = []
        for j in xrange(header.analog_in_channel_count):
            chan_d = AcquireChannelDescriptor(buf, loc_bytes)
            loc_bytes = chan_d.next_loc_bytes
            blocks = []
            for k in xrange(chan_d.block_count):
                block_desc = AcquireBlockDescriptor(buf, loc_bytes)
                loc_bytes = block_desc.next_loc_bytes
                blocks.append(block_desc)
            seg.channels.append(blocks)
        sweep.segments.append(seg)
    block.sweep = sweep

def ReadSeries(buf, loc_bytes, block, header):
    series = AcquireSeriesDescriptor(buf, loc_bytes)
    loc_bytes = series.next_loc_bytes
    ints = read_struct(buf, loc_bytes, '%iq'%series.sweep_count)
    series.sweep_dir = ints[:-1]
    block.series = series

def ReadBlock(buf, loc_bytes, header):
    tag = AcquireBlockTag(buf, loc_bytes)
    block = Block(tag)
    if tag.block_type == ACQUIRE_FILE_PREFIX_BLOCK:
        ReadFilePrefix(buf, tag.next_loc_bytes, block, header)
    elif tag.block_type == ACQUIRE_SWEEP_PREFIX_BLOCK:
        ReadSweepPrefix(buf, tag.next_loc_bytes, block, header)
    #elif tag.block_type == ACQUIRE_DATA_BLOCK:
    elif tag.block_type == ACQUIRE_SWEEP_BLOCK:
        ReadSweep(buf, tag.next_loc_bytes, block, header)
    elif tag.block_type == ACQUIRE_SERIES_BLOCK:
        ReadSeries(buf, tag.next_loc_bytes, block, header)
    #elif tag.block_type == ACQUIRE_CONTROL_BLOCK:
    return block, round_up(loc_bytes + tag.block_size, ACQUIRE_BLOCK_SIZE)
    
    
    


class AcquireFile(object):
    def __init__(self, path):
        size = os.stat(path)[6]
        self.fileno = os.open(path, os.O_RDONLY)
        self.mmap = mmap.mmap(self.fileno, size, access=mmap.ACCESS_READ)
        self.ai_channels = []
        self.series = []
        self.data_by_loc_bytes = {}
        self.header = None
        self.file_prefix = None
        self.control = None
        try:
            self.read_blocks()
        except:
            if not (self.header and self.file_prefix and self.series):
                raise
            else:
                traceback.print_exc()
        self.sampling_sec = self.header.frame_interval * 1.0 / self.header.clock_rate
    def __del__(self):
        self.close()
    def close(self):
        if self.mmap:
            self.mmap.close()
            self.mmap = None
            os.close(self.fileno)
    def read_blocks(self):
        self.header, loc_bytes = ReadHeader(self.mmap)
        #  collectors for this/next series:
        self.sweep_prefix = None
        self.sweeps = []
        while True:
            block, next_loc_bytes = ReadBlock(self.mmap, loc_bytes, self.header)
            if block.tag.block_type == ACQUIRE_FILE_PREFIX_BLOCK:
                self.file_prefix = block
                self.ai_channels = self.file_prefix.analog_in_channel_records
            elif block.tag.block_type == ACQUIRE_SWEEP_PREFIX_BLOCK:
                self.sweep_prefix = block
            elif block.tag.block_type == ACQUIRE_DATA_BLOCK:
                self.data_by_loc_bytes[loc_bytes] = block
            elif block.tag.block_type == ACQUIRE_SWEEP_BLOCK:
                self.sweeps.append(block)
                block.prefix = self.sweep_prefix
                self.sweep_prefix = None
            elif block.tag.block_type == ACQUIRE_SERIES_BLOCK:
                self.series.append(block)
                block.sweeps = self.sweeps
                self.sweeps = []
            elif block.tag.block_type == ACQUIRE_CONTROL_BLOCK:
                self.control = block
                break # no more useful blocks
            loc_bytes = next_loc_bytes
    def read_data_int16(self, series, sweep, segment, channel, offset, count):
        data = numpy.zeros(shape=(count,), dtype='int16')
        ser = self.series[series]
        swp = ser.sweeps[sweep]
        seg = swp.sweep.segments[segment]
        chn = seg.channels[channel]
        at = 0
        b = 0
        while b < len(chn):
            if offset < (at + chn[b].data_count):
                break
            else:
                at += chn[b].data_count
                b += 1
        into = 0
        while count and (b < len(chn)):
            start = offset - at
            count_here = min(count, chn[b].data_count - start)
            segdata = self.decode(chn[b])
            data[into:into+count_here] = segdata[start:start+count_here]
            into += count_here
            count -= count_here
            offset += count_here
            at += chn[b].data_count
            b += 1
        return data
    def read_data_float32(self, series, sweep, segment, channel, offset, count):
        ser = self.series[series]
        swp = ser.sweeps[sweep]
        ch_fi = self.file_prefix.analog_in_channel_records[channel]
        ch_pre = swp.prefix.analog_in_channel_prefixes[channel]
        idata = self.read_data_int16(series, sweep, segment, channel, offset, count)
        fdata = idata.astype('float32')
        fdata *= ( (1.0 / (ch_pre.gain * pow(1e3, ch_fi.natural_scale)))
                   * (ch_pre.maximum_voltage - ch_pre.minimum_voltage) / (ch_pre.maximum_value - ch_pre.minimum_value))
        return fdata
    def decode(self, block_desc):
        count = block_desc.data_count
        size = block_desc.data_size
        block = self.data_by_loc_bytes[block_desc.location]
        loc = block.tag.next_loc_bytes
        src = numpy.frombuffer(self.mmap, 'int8', size, loc)
        src_16 = numpy.frombuffer(self.mmap, numpy.dtype('>i2'), size/2, loc)
        src_16s = numpy.frombuffer(self.mmap, numpy.dtype('>i2'), size/2, loc+1)
        dest = numpy.zeros(shape=(count,), dtype='int16')
        decode_acquirefile_samples(src, src_16, src_16s, dest, size, count)
        return dest


# .acquire version _version_._minor_version_ by _vendor_
# _date_time_
# _comment_
# _technique_
# _user_name_ @ _computer_name_

# Series _i_:
#   _comment_
#   _technique_
#   [Sweep _j_: _comment_ / _technique_]



# segment relative dates
# channel nameL _label_ (_signal_)
# scaling and units

# qdf use scaling=scaling[0]; dcscaling[i] = scaling[i]/scaling[0]


def ScaleUnit(value, sUnit=''):
    """Returns a scaled value of "val" as function return with a single character unit
scaler from the metric scale set as a tuple of the form (Scaled Value, Scaler Char+sUnit)"""
    if value:
        SuffixIndex=int(floor(log10(abs(value))/3))
        if SuffixIndex!=0:
    
#Tack on the scale character. Note that SuffixIndex will be -1 for milli, -2
#for micro, etc.  Since 'm' is the 6th character in the suffix array, we need
#to add 6 to get -1 (for milli) to point at the 6th character. It also means that
#we must limit the index to be between -6 (+6=0, the first character) and 4
#(+6=10, the last character).
    
            SuffixIndex = EnsureRange(SuffixIndex,-6,4)
            return (value*10**(-3*SuffixIndex),'afpnum%KMGT'[SuffixIndex+6]+sUnit);
    return (value,sUnit)



if __name__ == '__main__':
    import sys
    from itertools import *

    f = AcquireFile(sys.argv[1])
    for r, series in enumerate(f.series):
        #print 'Series %i:' % (r+1)
        for w, sweep in enumerate(series.sweeps):
            #print 'Sweep %i:' % (w+1)
            for g, seg in enumerate(sweep.sweep.segments):
                #print 'Segment %i:' % (g+1)
                at = 0
                for chans in izip(*seg.channels):
                    bufs = [f.read_data_float32(r, w, g, c, at, block.data_count) for c, block in enumerate(chans)]
                    #bufs = [f.read_data_int16(r, w, g, c, at, block.data_count) for c, block in enumerate(chans)]
                    at += len(bufs[0])
                    for nums in izip(*bufs):
                        print '\t'.join([str(x) for x in nums])
                print



